# Methods to calculate WCAG contrast ratio and check if it passes AA or AAA
# https://www.w3.org/WAI/WCAG21/Understanding/contrast-minimum.html

import re

from typing import NewType, Tuple, TypeAlias, Union


# float01 is a float greater than or equal to 0 and less than or equal to 1
float01 = NewType("float01", float)
RGBColor: TypeAlias = Tuple[float01, float01, float01]


def hexdd_to_float01(xx: str) -> float01:
    """Convert a two-digit, 8-bit hex value to a float between 0 and 1.

    >>> hexdd_to_float01("00")
    0.0
    >>> hexdd_to_float01("33")
    0.2
    >>> hexdd_to_float01("55")
    0.3333333333333333
    >>> hexdd_to_float01("cc")
    0.8
    >>> hexdd_to_float01("ff")
    1.0
    """
    return float01(int(xx, 16) / 255)


def hex_to_rgb01(hex: str) -> RGBColor:
    """Convert a hex defined colour to RGB.

    Args:
        hex (string): color in hex format (#rrggbb/#rgb)

    Returns:
        rgb: tuple of rgb values ``(r, g, b)``, where each channel (red, green,
        blue) can assume values between 0 and 1 (inclusive).
    """

    if re.match(r"\A#[a-fA-F0-9]{6}\Z", hex):
        # Full hex color (#rrggbb) format
        return (
            hexdd_to_float01(hex[1:3]),
            hexdd_to_float01(hex[3:5]),
            hexdd_to_float01(hex[5:7]),
        )

    if re.match(r"\A#[a-fA-F0-9]{3}\Z", hex):
        # Short hex color (#rgb) format, shorthand for #rrggbb
        return (
            hexdd_to_float01(hex[1] * 2),
            hexdd_to_float01(hex[2] * 2),
            hexdd_to_float01(hex[3] * 2),
        )

    raise ValueError("Invalid hex color format")


def sRGB_channel(v: float01) -> float:
    """Colors need to be normalised (from a sRGB space) before computing the relative
    luminance.

    Args:
        v (float): r,g,b channels values between 0 and 1

    Returns:
        float: sRGB channel value for a given rgb color channel
    """
    if v <= 0.04045:
        return v / 12.92
    else:
        return ((v + 0.055) / 1.055) ** 2.4


def relative_luminance(color: RGBColor) -> float:
    """Compute the relative luminance of a color.

    Args:
        color (tuple): rgb color tuple ``(r, g, b)``

    Returns:
        float: relative luminance of a color
    """
    r, g, b = color
    r_ = sRGB_channel(r)
    g_ = sRGB_channel(g)
    b_ = sRGB_channel(b)

    return 0.2126 * r_ + 0.7152 * g_ + 0.0722 * b_


def contrast_ratio(color1: RGBColor, color2: RGBColor) -> float:
    """Compute the contrast ratio between two colors.

    Args:
        color1 (tuple): rgb color tuple ``(r, g, b)``
        color2 (tuple): rgb color tuple ``(r, g, b)``

    Returns:
        float: contrast ratio between two colors
    """

    l1 = relative_luminance(color1)
    l2 = relative_luminance(color2)

    if l1 > l2:
        return (l1 + 0.05) / (l2 + 0.05)
    else:
        return (l2 + 0.05) / (l1 + 0.05)


def passes_contrast(
    color1: RGBColor, color2: RGBColor, level="AA"
) -> Union[bool, float]:
    """Method to verify the contrast ratio between two colours.

    Args:
        color1 (tuple): rgb color tuple ``(r, g, b)``
        color2 (tuple): rgb color tuple ``(r, g, b)``
        level (str, optional): WCAG contrast level. Defaults to "AA".
    """

    ratio = contrast_ratio(color1, color2)

    if level == "AA":
        if ratio >= 4.5:
            return round(ratio, 2)
        else:
            return False
    elif level == "AAA":
        if ratio >= 7.0:
            return round(ratio, 2)
        else:
            return False

    raise ValueError("level must be either 'AA' or 'AAA'")


def get_wcag_level_normal_text(contrast: float) -> str:
    """Does the given contrast meet level AA or level AAA for normal size text"""
    if contrast >= 7:
        return "AAA"
    elif contrast >= 4.5:
        return "AA"
    else:
        return ""


def get_wcag_level_large_text(contrast: float) -> str:
    """Does the given contrast meet level AA or level AAA for large text"""
    if contrast >= 4.5:
        return "AAA"
    elif contrast >= 3:
        return "AA"
    else:
        return ""


def hexstr_without_hash(hex_color: str) -> str:
    """Remove '#' from hex color strings.

    Usage example:

    >>> hexstr_without_hash("#fff")
    'fff'
    """
    return hex_color.replace("#", "")



from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    Other,
    Punctuation,
    String,
    Text,
)


class Colors:
    comment = "#515151"
    red = "#d71835"
    orange = "#7f4707"
    yellow = "#7f4707"
    green = "#116633"
    blue = "#00749c"
    purple = "#8045e5"
    black = "#1e1e1e"


class Theme(Style):
    """
    This style inspired by the a11y-light theme from eric bailey's accessible themes.
    """

    default_style = ""

    background_color = "#f2f2f2"
    highlight_color = "#fdf2e2"

    styles = {
        Text: Colors.black,  # class:  ''
        Error: Colors.red,  # class: 'err'
        Other: "",  # class 'x'
        Comment: Colors.comment,  # class: 'c'
        Keyword: Colors.purple,  # class: 'k'
        Keyword.Constant: Colors.purple,  # class: 'kc'
        # Keyword.Declaration:       "",            # class: 'kd'
        # Keyword.Namespace:         "",            # class: 'kn'
        # Keyword.Pseudo:            "",            # class: 'kp'
        # Keyword.Reserved:          "",            # class: 'kr'
        Keyword.Type: Colors.orange,  # class: 'kt'
        Operator: Colors.green,  # class: 'o'
        Operator.Word: Colors.purple,  # class: 'ow'
        Punctuation: Colors.black,  # class: 'p'
        Name: Colors.black,  # class: 'n'
        Name.Attribute: Colors.yellow,  # class: 'na'
        Name.Builtin: Colors.orange,  # class: 'nb'
        Name.Builtin.Pseudo: Colors.orange,  # class: 'bp'
        Name.Class: Colors.blue,  # class: 'nc'
        Name.Constant: Colors.blue,  # class: 'no'
        Name.Decorator: Colors.orange,  # class: 'nd'
        Name.Entity: Colors.green,  # class: 'ni'
        Name.Exception: Colors.purple,  # class: 'ne'
        Name.Function: Colors.blue,  # class: 'nf'
        Name.Property: Colors.blue,  # class: 'py'
        Name.Label: Colors.orange,  # class: 'nl'
        Name.Namespace: Colors.black,  # class: 'nn'
        # Name.Other:                "",            # class: 'nx'
        Name.Tag: Colors.blue,  # class: 'nt'
        Name.Variable: Colors.red,  # class: 'nv'
        Name.Variable.Magic: Colors.orange,
        # Name.Variable.Class:       "",            # class: 'vc'
        # Name.Variable.Global:      "",            # class: 'vg'
        # Name.Variable.Instance:    "",            # class: 'vi'
        Number: Colors.orange,  # class: 'm'
        # Number.Float:              "",            # class: 'mf'
        # Number.Hex:                "",            # class: 'mh'
        # Number.Integer:            "",            # class: 'mi'
        # Number.Integer.Long:       "",            # class: 'il'
        # Number.Oct:                "",            # class: 'mo'
        Literal: Colors.orange,  # class: 'l'
        # Literal.Date:              "",            # class: 'ld'
        String: Colors.green,  # class: 's'
        String.Backtick: Colors.green,  # class: 'sb'
        # String.Char:               "",            # class: 'sc'
        # String.Doc:                "",            # class: 'sd'
        # String.Double:             "",            # class: 's2'
        # String.Escape:             "",            # class: 'se'
        # String.Heredoc:            "",            # class: 'sh'
        # String.Interpol:           "",            # class: 'si'
        # String.Other:              "",            # class: 'sx'
        String.Regex: Colors.red,  # class: 'sr'
        # String.Single:             "",            # class: 's1'
        String.Symbol: Colors.blue,  # class: 'ss'
        # Generic:                   "",            # class: 'g'
        Generic.Deleted: Colors.blue,  # class: 'gd',
        Generic.Emph: "italic",  # class: 'ge'
        # Generic.Error:             "",            # class: 'gr'
        Generic.Heading: Colors.blue,  # class: 'gh'
        Generic.Subheading: Colors.blue,  # class: 'gu'
        # Generic.Inserted:          "",            # class: 'gi'
        # Generic.Output:            "",            # class: 'go'
        # Generic.Prompt:            "",            # class: 'gp'
        Generic.Strong: "bold",  # class: 'gs'
        # Generic.Traceback:         "",            # class: 'gt'
    }



import logging
import os

from pathlib import Path
from typing import Union

from pygments.formatters import HtmlFormatter
from pygments.styles import get_style_by_name
from pygments.token import Text
from setuptools import find_packages


def find_all_themes_packages() -> list[str]:
    """Finds the currently supported themes in the a11y_pygments package.

    Returns:
        themes: list of themes under the a11y_pygments package
    """
    exclude = {"test", "scripts", "a11y_pygments", "a11y_pygments.utils"}
    packages = set(find_packages())
    themes = list(packages - exclude)
    # drop the a11y_pygments part of the pkg name
    themes = [x.split(".")[1] for x in themes]
    return themes


def get_themes_names() -> list[str]:
    """Get themes names from the a11y_pygments package.

    Returns:
        themes: list of themes names
    """
    themes = find_all_themes_packages()
    themes = [x.replace("_", "-") for x in themes]
    logging.info(f"Found pygment themes: {themes}")
    return themes


def generate_css(themes: list[str], save_dir: Union[str, Path]):
    """Generate css for the available themes.
    Args:
        themes (list): list of themes names
    """
    assert save_dir, "Must provide directory"

    for theme in themes:
        style = get_style_by_name(theme)
        formatter = HtmlFormatter(style=style, full=True, hl_lines=[2, 3, 4])
        css = formatter.get_style_defs()
        color = style.style_for_token(Text)["color"]
        css += (
            f"\n.highlight {{ background: {style.background_color}; color: #{color}; }}"
        )
        package = theme.replace("-", "_")

        if not Path(save_dir).joinpath("css").exists():
            os.mkdir(Path(save_dir).joinpath("css"))
        if "docs" in save_dir:
            out = Path(save_dir).joinpath("css", f"{package}-style.css")
        else:
            out = Path(save_dir) / package / "style.css"

        logging.info(f"Saving css to {out}")
        with open(out, "w") as f:
            f.write(css)
